package edu.cmu.cs.sasylf.ast;


import java.util.*;
import java.io.*;

import edu.cmu.cs.sasylf.term.*;
import edu.cmu.cs.sasylf.util.ErrorHandler;

import static edu.cmu.cs.sasylf.util.Util.*;
import static edu.cmu.cs.sasylf.ast.Errors.*;


public abstract class DerivationByAnalysis extends Derivation {
	public DerivationByAnalysis(String n, Location l, Clause c, String derivName) {
		super(n,l,c); targetDerivationName = derivName;
	}

	public String getTargetDerivationName() { return targetDerivationName; }

	protected void computeTargetDerivation(Context ctx) {
		if (targetDerivation == null) {
			targetDerivation = ctx.derivationMap.get(targetDerivationName);
			debug("targetDerivation is " + targetDerivation);
			if (targetDerivation == null) {
				for (FreeVar v : ctx.inputVars) {
					if (targetDerivationName.equals(v.getName())) {
						if (this instanceof DerivationByInduction)
							ErrorHandler.report("Cannot perform induction over "+ targetDerivationName + " unless you add this variable as a specific \"forall\" clause of the theorem", this);
						else {
							targetDerivation = new SyntaxAssumption(targetDerivationName, getLocation());
							targetDerivation.typecheck(ctx);
							return;
						}
					}
				}
				ErrorHandler.report(DERIVATION_NOT_FOUND, "Cannot find a derivation named "+ targetDerivationName, this);
			}
		}
	}
	public Fact getTargetDerivation() {
		return targetDerivation;
	}

	public List<Case> getCases() { return cases; }

	public abstract String byPhrase();

	public void prettyPrint(PrintWriter out) {
		super.prettyPrint(out);
		out.println(" by "+byPhrase()+" on " + targetDerivationName + ":");
		for (Case c: cases) {
			c.prettyPrint(out);
		}
		out.println("end " + byPhrase());
	}

	public void typecheck(Context ctx) {
		computeTargetDerivation(ctx);

		super.typecheck(ctx);

		Term oldCase = ctx.currentCaseAnalysis;
		Element oldElement = ctx.currentCaseAnalysisElement;
		Term oldGoal = ctx.currentGoal;
		Map<CanBeCase,Set<Pair<Term,Substitution>>> oldCaseTermMap = ctx.caseTermMap;
		ctx.currentCaseAnalysis = adapt(targetDerivation.getElement().asTerm(), targetDerivation.getElement(), ctx);
		//ctx.currentCaseAnalysis = targetDerivation.getElement().asTerm().substitute(ctx.currentSub);
		ctx.currentCaseAnalysisElement = targetDerivation.getElement();
		ctx.currentGoal = getElement().asTerm().substitute(ctx.currentSub);
		
		boolean isSubderivation = targetDerivation != null && targetDerivation.equals(ctx.inductionVariable);
		if (isSubderivation) debug("found subderivation: " + targetDerivation);
		
		ctx.caseTermMap = new HashMap<CanBeCase,Set<Pair<Term,Substitution>>>();

		if (ctx.currentCaseAnalysisElement instanceof NonTerminal) {
			Syntax syntax = ((NonTerminal)ctx.currentCaseAnalysisElement).getType();

			for (Clause clause : syntax.getClauses()) {
				if (clause.isVarOnlyClause()) {
					continue; // no cases for var clauses
				}
				Term term = ((ClauseDef)clause).getSampleTerm();
				Substitution freshSub = term.freshSubstitution(new Substitution());
				term.substitute(freshSub);
				Set<Pair<Term,Substitution>> set = new HashSet<Pair<Term,Substitution>>();
				set.add(new Pair<Term,Substitution>(term, new Substitution()));
				ctx.caseTermMap.put(clause, set);
			}
			
		} else {
			Judgment judge= (Judgment) ((ClauseUse)ctx.currentCaseAnalysisElement).getConstructor().getType();
			debug("*********** case analyzing line " + getLocation().getLine());
			//debug("    sub = " + ctx.currentSub);
			//debug("    adaptationSub = " + ctx.adaptationSub);
			// see if each rule, in turn, applies
			for (Rule rule : judge.getRules()) {
				Set<Pair<Term,Substitution>> caseResult = rule.caseAnalyze(ctx);
				ctx.caseTermMap.put(rule, caseResult);
			}
		}

		
		for (Case c : cases) {
			c.typecheck(ctx, isSubderivation);
		}
		
		for (Map.Entry<CanBeCase, Set<Pair<Term,Substitution>>> entry : ctx.caseTermMap.entrySet()) {
			if (!entry.getValue().isEmpty()) {
				CanBeCase cbc = entry.getKey();
				ErrorHandler.report(Errors.MISSING_CASE,
									cbc.getErrorDescription()/* + ": case is " + ctx.currentCaseAnalysis.substitute(sub)
									+ "\n\trule conc was " + arguments.get(lastIdx)*/, this);				
			}
		}
		
		/*if (!(ctx.currentCaseAnalysisElement instanceof NonTerminal) && !ctx.caseTermMap.isEmpty()) {
			Rule rule = ctx.caseTermMap.keySet().iterator().next();
			ErrorHandler.report(Errors.MISSING_CASE,
								rule.getName()/* + ": case is " + ctx.currentCaseAnalysis.substitute(sub)
								+ "\n\trule conc was " + arguments.get(lastIdx)* /, this);
		}*/

		ctx.caseTermMap = oldCaseTermMap;
		ctx.currentCaseAnalysis = oldCase;
		ctx.currentCaseAnalysisElement = oldElement;
		ctx.currentGoal = oldGoal ;
	}

	/** Adapts this term using the current context
	 * This includes substituting with the current sub
	 * and also adapting the context to include assumptions currently in scope
	 */
	protected static Term adapt(Term term, Element element, Context ctx) {
		// TODO: generalize this to all terms reference in system
		if (ctx.adaptationSub != null && term.countLambdas() < ctx.matchTermForAdaptation.countLambdas() && element instanceof ClauseUse) {
			term = ((ClauseUse)element).adaptTermTo(term, ctx.matchTermForAdaptation, ctx.adaptationSub, true);
		}
		
		return term.substitute(ctx.currentSub);
	}

	private List<Case> cases = new ArrayList<Case>();
	private String targetDerivationName;
	private Fact targetDerivation;
}
